#include "stdafx.h"
#include "Field.h"
#include "Site.h"
#include "Context.h"
#include "Space.h"
#include "MASS.h"
#include "MMASS.h"
using namespace MMASS;

Site::~Site() {
}

const string& Site::getName() const {
	return name;
}

const spSite Site::getSelf() const {
	return self.lock();
}

const bool Site::hasAdjacentSites() const {
	return (!adjSites.empty());
}

const bool Site::isAdjacentTo(const string &name_) const {
	return (adjSites.count(name_) > 0);
}

const bool Site::isAdjacentTo(const spSite site) const {
	return (this->isAdjacentTo(site->getName()));
}

const size_t Site::numOfAdjacentSites() const {
	return adjSites.size();
}

void Site::addAdjacentSite(const spSite site) throw (invalid_argument) {
	if (!site)
		throw invalid_argument("Site::addAdjacentSite(const spSite site) - "
			"Parameter 'site' doesn't point to a site (NULL Site). A site "
			"can't be adjacent to a null site.\n");

	if (this->isAdjacentTo(site)) return;

	if (site->getSelf() == self.lock())
		throw invalid_argument("Site::addAdjacentSite(const spSite site) - "
			" site->getName(): "+site->getName()+"\n"
			"You are trying to make site "+site->getName()+" adjacent to"
			"site \""+name+"\". A site can't be adjacent to"
			" itself.\n");

	if (site->getSpace()->getMASS() != space.lock()->getMASS())
		throw invalid_argument("Site::addAdjacentSite(const spSite site) - "
			" site->getName(): "+site->getName()+"\n"
			"Site \""+site->getName()+"\" is in a different space of site \""
			+name+"\" (MASSs' name are "
			+site->getSpace()->getMASS()->getName()+" and "
			+space.lock()->getMASS()->getName()+" respectively). A site "
			"can't be adjacent to a site of another space.\n");

	adjSites[site->getName()] = site;
	try {
		site->addAdjacentSite(self.lock());
	} catch (invalid_argument &e) {
		throw invalid_argument(string("Site::addAdjacentSite(const spSite site) - "
			"\n<-- ")+e.what());
	}
}

void Site::addAdjacentSites(const spSite_set sites) throw (invalid_argument) {
	spSite nullSite;
	spSite_set::const_iterator i = sites.find(nullSite);
	if (i != sites.end())
		throw invalid_argument("Site::addAdjacentSites(const spSite_set sites) -"
			" sites has a null site. A site can't be adjacent to a null site.\n");

	spSite_set::const_iterator end = sites.end();
	for (i = sites.begin(); i != end; ++i) {
		if ((*i)->getSpace() != space.lock())
			throw invalid_argument("Site::addAdjacentSites(const spSite_set sites) -"
				" The site \""+(*i)->getName()+"\" in 'sites' is in space from MASS \""
				+(*i)->getSpace()->getMASS()->getName()+"\", this space is from MASS \""
				+space.lock()->getMASS()->getName()+"\". A site can't be adjacent to"
				" a site in a different space.\n");
	}

	try {
		for (i = sites.begin(); i != end; ++i)
			this->addAdjacentSite((*i));
	} catch (invalid_argument &e) {
		throw invalid_argument(string("Site::addAdjacentSites(const spSite_set sites) - "
			"\n<-- ")+e.what());
	}
}

wpSite_map::const_iterator Site::beginAdjacentSites() const {
	return adjSites.begin();
}

wpSite_map::const_iterator Site::endAdjacentSites() const {
	return adjSites.end();
}

queue<const string> Site::getSitesInDistanceOrder() {
	return sitesInDistanceOrder;
}

const unsigned long Site::getDistanceOf(spSite site) {
	// OR THROW A EXCEPTION!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! <----------------
	if (sitesDistances.count(site->getName()) == 0)
		return 0;

	return sitesDistances[site->getName()];
}

const bool Site::hasAgent() const {
	return agent.lock();
}

spAgent Site::getAgent() const {
	return agent.lock();
}

void Site::setAgent(spAgent agent_) throw (logic_error) {
	if (!agent_)
		throw logic_error("Site::setAgent(spAgent agent_) - "
			"agent_: null\n"
			"Can't set a site's agent to null. (site: \""
			+name+"\".\n");

	spAgent a = agent.lock();
	if (a) {
		if (a == agent_) return;

		throw logic_error("Site::setAgent(spAgent agent_) - "
			"site_: "+agent_->getName()+"\n"
			"This site already has the agent \""
			+agent.lock()->getName()+"\". Can't set the site \""
			+name+"\"'s agent.\n");
	}

	if (agent_->getSite() != self.lock()) {
		try {
			agent_->setSite(self.lock());
		} catch (Site::logic_error &e) {
			throw logic_error(string("Agent::setSite(spSite site_) -\n"
				"<-- ")+e.what());
		}
	}

	agent = agent_;
	if (agent_->getSite() == self.lock()) return;

	try {
		agent_->setSite(self.lock());
	} catch (Site::logic_error &e) {
		throw logic_error(string("Agent::setSite(spSite site_) -\n"
			"<-- ")+e.what());
	}
}

const spSpace Site::getSpace() const {
	return space.lock();
}

Site::Site(const string &name_) : name(name_) {
	visited = false;
}

const spSite Site::createSite(const string &name_) {
	spSite site(new Site(name_));
	site->setSelfPointer(site);

	return site;
}

spContext Site::createContext() {
	spSiteInfo_set adjSitesInfo;
	spAgentInfo_set adjAgentsInfo;

	wpSite_map::const_iterator site_i, end = this->adjSites.end();
	for (site_i = this->adjSites.begin(); site_i != end; ++site_i) {
		spSite site = (*site_i).second.lock();
		adjSitesInfo.insert(site->createSiteInfo());

		spAgent agent = site->getAgent();
		if (agent)
			adjAgentsInfo.insert(agent->createAgentInfo());
	}

	return Context::createContext(this->createSiteInfo(), adjSitesInfo, adjAgentsInfo);
}

void Site::setSelfPointer(const wpSite self_) {
	self = self_;
}

void Site::setSpace(wpSpace space_) {
	space = space_;
}

void Site::setAgentToNull() {
	spAgent a = agent.lock();
	if (!a) return;
	agent.reset();
	a->setSiteToNull();

}

void Site::addField(spField field, double value) {
	const string id = field->getIdentifier();
	if (fields.count(id) > 0) {
		double activeValue = fields[id]->getValue();
		fields[id]->setValue(fields[id]->compose(activeValue, value));
	} else {
		fields[id] = field->createField(value);
	}
	//field.reset();
}

void Site::clearFields() {
	//printf("%lu - clearing %s\n", this->space.lock()->getMASS()->getMMASS()->getTurn(), this->name.c_str()); fflush(NULL);
	spField_map::iterator field_i, f_end = this->fields.end();
	for (field_i = this->fields.begin(); field_i != f_end; ++field_i) {
		(*field_i).second->setValue(0);
	}
	//this->fields.clear();
}

void Site::setVisited(bool b) {
	visited = b;
}

const bool Site::wasVisited() const {
	return visited;
}

void Site::calculateDistances() {
	queue<wpSite> sitesQueue;
	sitesQueue.push(this->self);

	spSite_map::const_iterator s_i, s_e = this->space.lock()->endSites();
	for (s_i = this->space.lock()->endSites(); s_i != s_e; ++s_i)
		this->sitesDistances[(*s_i).second->getName()] = 0;

	while(!sitesQueue.empty()) {
		wpSite s = sitesQueue.front();
		sitesQueue.pop();

		wpSite_map::const_iterator site_i, s_end = s.lock()->adjSites.end();
		for (site_i = s.lock()->adjSites.begin(); site_i != s_end; ++site_i) {
			const string name = (*site_i).second.lock()->getName();
			if ((name != this->name) && (this->sitesDistances[name] == 0)) {
				sitesQueue.push((*site_i).second);
				this->sitesDistances[name] = this->sitesDistances[s.lock()->getName()] + 1;
				this->sitesInDistanceOrder.push(name);
			}
		}
	}
}

spSiteInfo Site::createSiteInfo() {
	spField_set fieldsSet;

	if (!this->fields.empty()) {
	spField_map::iterator field_i, f_end = this->fields.end();
	//printf("%lu - accessing %s\n", this->space.lock()->getMASS()->getMMASS()->getTurn(), this->name.c_str()); fflush(NULL);
	for (field_i = this->fields.begin(); field_i != f_end; ++field_i)
		fieldsSet.insert((*field_i).second);
	}

	if (agent.lock())
		return SiteInfo::createSiteInfo(agent.lock()->getName(), this->name, fieldsSet);
	return SiteInfo::createSiteInfo(this->name, fieldsSet);
}

spContext Site::sense(spAgent agent) {
	return spContext(agent->getSite()->createContext());
}

//bool Site::act(spAgent agent, unsigned long turn_) {
	//return this->space.lock()->getMASS()->getMMASS()->act(agent, turn_);
//}

const string& SiteInfo::getAgentName() { return agentName; }
const string& SiteInfo::getSiteName() { return siteName; }

spField_set::const_iterator SiteInfo::beginFields() { return fields.begin(); }
spField_set::const_iterator SiteInfo::endFields() { return fields.end(); }

SiteInfo::SiteInfo(const string &siteName_, spField_set &fields_) : siteName(siteName_), fields(fields_) { has_agent = false; }
SiteInfo::SiteInfo(const string &agentName_, const string &siteName_, spField_set &fields_) : agentName(agentName_), siteName(siteName_), fields(fields_) { has_agent = true; }
spSiteInfo SiteInfo::createSiteInfo(const string &siteName_, spField_set fields_) { return spSiteInfo(new SiteInfo(siteName_, fields_)); }
spSiteInfo SiteInfo::createSiteInfo(const string &agentName_, const string &siteName_, spField_set fields_) { return spSiteInfo(new SiteInfo(agentName_, siteName_, fields_)); }
