#ifndef SITE_H
#define SITE_H

#ifdef MMASSDLL_EXPORTS
#define MMASSDLL_API __declspec(dllexport)
#else
#define MMASSDLL_API __declspec(dllimport)
#endif

#include "MMASSDefs.h"
#include <OgreNoMemoryMacros.h>
#include <string>
#include <stdexcept>
#include <queue>
#include <map>

namespace MMASS {
	using std::string;
	using std::map;
	using std::queue;

	class MMASSDLL_API Site {
	public:
		class logic_error : public std::logic_error {
		public:
			logic_error(const string& msg = "") : std::logic_error(msg) {}
		};

		class invalid_argument : public std::invalid_argument {
		public:
			invalid_argument(const string& msg = "") : std::invalid_argument(msg) {}
		};

	protected:
		wpAgent     agent;    // agent
		wpSite_map  adjSites; // adjacent sites
		spField_map fields;   // fields

	public:
		spContext createContext(); // should be private

		~Site();

		const string& getName() const;

		const spSite getSelf() const;

		const bool hasAdjacentSites() const;
		const bool isAdjacentTo(const string &name_) const;
		const bool isAdjacentTo(const spSite site) const;
		const size_t numOfAdjacentSites() const;
		void addAdjacentSite(const spSite site) throw (invalid_argument);
		void addAdjacentSites(const spSite_set sites) throw (invalid_argument);
		wpSite_map::const_iterator beginAdjacentSites() const;
		wpSite_map::const_iterator endAdjacentSites() const;
		queue<const string> getSitesInDistanceOrder();
		const unsigned long getDistanceOf(spSite site);

		const bool hasAgent() const;
		spAgent getAgent() const;
		void setAgent(spAgent agent_) throw (logic_error);

		const spSpace getSpace() const;

	private:
		const string                     name;
		wpSite                           self;
		bool                             visited;
		map<const string, unsigned long> sitesDistances;
		queue<const string>              sitesInDistanceOrder;
		wpSpace                          space;

		Site(const string &name_);

		static const spSite createSite(const string &name_);

		// spContext createContext();

		void setSelfPointer(const wpSite self_);

		void setSpace(wpSpace space_);

		void setAgentToNull();

		void addField(spField field, double value);
		void clearFields();

		void setVisited(bool b);
		const bool wasVisited() const;

		void calculateDistances();

		spSiteInfo createSiteInfo();

		spContext sense(spAgent agent);

		//bool act(spAgent agent, unsigned long turn_);

		/* Hiding methods. */
		Site(const Site&);
		Site& operator=(const Site&);

		friend class Transport;
		friend class Emit;
		friend class Agent;
		friend class Space;
		// friend MMASSDLL_API void* _agentBehaviourThread(void *arg);
		friend MMASSDLL_API void _clearFields(spMMASS mmass);

		friend MMASSDLL_API void printSitesState(spMMASS mmass);
	};

	class MMASSDLL_API SiteInfo {
	public:
		const string& getAgentName();
		const string& getSiteName();

		spField_set::const_iterator beginFields();
		spField_set::const_iterator endFields();

	protected:
		const string agentName;
		const string siteName;
		spField_set  fields;
		bool         has_agent;

	private:
		SiteInfo(const string &siteName_, spField_set &fields_);
		SiteInfo(const string &agentName_, const string &siteName_, spField_set &fields_);
		static spSiteInfo createSiteInfo(const string &siteName_, spField_set fields_);
		static spSiteInfo createSiteInfo(const string &agentName_, const string &siteName_, spField_set fields_);

		friend class Site;
	};
}

#endif