/*
-----------------------------------------------------------------------------
This source file is part of OGRE
(Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2005 The OGRE Team
Also see acknowledgements in Readme.html

You may use this sample code for anything you like, it is not covered by the
LGPL like the rest of the engine.
-----------------------------------------------------------------------------
*/


#include "ShepherdsDog.h"
#include <pthread.h>
#include <OgreNoMemoryMacros.h>
#include "../../MMASSdll/MMASSDefs.h"
#include "../../MMASSdll/Site.h"
#include "../../MMASSdll/Space.h"
#include "../../MMASSdll/MASS.h"
#include "../../MMASSdll/MMASS.h"
#include "../Sheep.h"
#include <deque>
#include <map>
#include <string>
#define HEIGHT 200
#define WIDTH 200
#define NUMROWS 20
#define NUMCOLS 20
using std::deque;
using std::map;
using std::string;

class MovingEntity {
public:
	MovingEntity() : mEntity(NULL), mNode(NULL) {
		mmass = MMASS::CMMASS::getSingleton();
		mWalking = false;
	}

	~MovingEntity() { }

public:
	Real mDistance;
	Vector3 mDirection;
	Vector3 mDestination;

	Entity *mEntity;
	SceneNode *mNode;
	bool mWalking;
	boost::shared_ptr< shepherdsdog::StateSheep > state;

	const static Real mWalkSpeed;

	MMASS::spMMASS mmass;
};
const Real MovingEntity::mWalkSpeed = 35.0f;

class MoveDemoListener : public ExampleFrameListener {
public:
	MoveDemoListener(RenderWindow *win, Camera *cam, map<const string, MovingEntity> &ents) 
		: ExampleFrameListener (win, cam, false, false), entities(ents) {
			fp = fopen("outputogre.txt", "w+b");
			mmass = MMASS::CMMASS::getSingleton();
			timeElapsed = 0.0f;
			turn = 0;
	}
	~MoveDemoListener() { fclose(fp); }

	bool nextLocation(const string entName) {
		if ( entities[entName].state->empty() )
            return false;

		entities[entName].mDestination = entities[entName].state->getNext( );

        entities[entName].mDirection = entities[entName].mDestination - entities[entName].mNode->getPosition( );
		entities[entName].mWalking = true;
        entities[entName].mDistance = entities[entName].mDirection.normalise( );

		return true;
	}

	bool frameStarted(const FrameEvent &evt) {
		using namespace MMASS;

		map<const string, MovingEntity>::iterator ent_i, ent_end = entities.end();
		for (ent_i = entities.begin(); ent_i != ent_end; ++ent_i) {
			const string entName = (*ent_i).first;
			if (entities[entName].mWalking || !nextLocation(entName)) {
				Real move = entities[entName].mWalkSpeed * evt.timeSinceLastFrame;
				entities[entName].mDistance -= move;

				if (entities[entName].mDistance <= 0.0f) {
					entities[entName].mNode->setPosition( entities[entName].mDestination );
					entities[entName].mWalking = false;
				} else {
					entities[entName].mNode->translate( entities[entName].mDirection * move );
				} 
			}
		}

		//fprintf(fp, "before if %lu\n", mmass->getTurn()); fflush(fp);
		//pthread_mutex_lock(&const_cast<pthread_mutex_t>(mmass->getOgreMutex()));
		//if (((mmass->getTurn() % 2) == 0) /*|| (mmass->waitOgre())*/) {
			//mmass->setWaitOgre(false);
			//pthread_cond_signal(&const_cast<pthread_cond_t>(mmass->getOgreCond()));
			//fprintf(fp, "ogre signal mmass\n"); fflush(fp);
		//}
		//pthread_mutex_unlock(&const_cast<pthread_mutex_t>(mmass->getOgreMutex()));
		//fprintf(fp, "after if %lu\n", mmass->getTurn()); fflush(fp);
		
		timeElapsed += evt.timeSinceLastFrame;
		if (timeElapsed >= timePerTurn) {
			pthread_mutex_lock(&const_cast<pthread_mutex_t>(mmass->getMutex()));
			mmass->nextOgreTurn();
			if (mmass->getTurn() <= (mmass->getOgreTurn()+ 3))
				pthread_cond_signal(&const_cast<pthread_cond_t>(mmass->getCond()));
			pthread_mutex_unlock(&const_cast<pthread_mutex_t>(mmass->getMutex()));
			timeElapsed = 0.0f;
		}

		bool mContinue = ExampleFrameListener::frameStarted(evt);

		if (!mContinue) {
			//pthread_mutex_lock(&const_cast<pthread_mutex_t>(mmass->getOgreMutex()));
			pthread_mutex_lock(&const_cast<pthread_mutex_t>(mmass->getMutex()));
			mmass->setFinishThread(true);
			pthread_mutex_unlock(&const_cast<pthread_mutex_t>(mmass->getMutex()));
			//pthread_cond_signal(&const_cast<pthread_cond_t>(mmass->getOgreCond()));
			//pthread_mutex_unlock(&const_cast<pthread_mutex_t>(mmass->getOgreMutex()));

			//std::map<const std::string, pthread_t>::const_iterator thread_i,
				//t_end = mmass->endAgentThreads();

			//for (thread_i = mmass->beginAgentThreads(); thread_i != t_end; ++thread_i)
				//pthread_join((*thread_i).second, NULL);
			//pthread_join(*(const_cast<pthread_t*>(&mmass->getEnvironmentThread())), NULL);
		}
		return mContinue;
	}

protected:
	map<const string, MovingEntity> entities;
	MMASS::spMMASS mmass;
	Real timeElapsed;
	FILE *fp;

	unsigned long turn;
	static const Real timePerTurn;
};
const Real MoveDemoListener::timePerTurn = 1.0f;

const std::string _getSiteName(const Ogre::Vector3 &v) {
	char name[6];
	int x = v.x;
	int z = v.z;
	int i = (x + 5) / 10;
	int j = (z + 5) / 10;
	sprintf(name,"%dx%d", i, j);

	return string(name);
}

class MoveDemoApplication : public ExampleApplication {
protected:
public:
	MoveDemoApplication() {
	}

	~MoveDemoApplication() {
	}

protected:
	map<const string, MovingEntity> entities;

	void createScene(void) {
		using namespace MMASS;
		using namespace shepherdsdog;

		mSceneMgr->setAmbientLight(ColourValue(1,1,1));
		mSceneMgr->setSkyDome( true, "Examples/CloudySky", 5, 8 );

		Plane plane(Vector3::UNIT_Y, 0);
		MeshManager::getSingleton().createPlane("ground",
			ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME, plane,
			NUMROWS*10,NUMCOLS*10,20,20,true,1,5,5,Vector3::UNIT_Z);

		Entity *ground = mSceneMgr->createEntity("GroundEntity", "ground");
		mSceneMgr->getRootSceneNode()->createChildSceneNode("groundnode", Vector3((NUMROWS*10)/2,0,(NUMCOLS*10)/2))->attachObject(ground);
		ground->setMaterialName("Examples/GrassFloor");
		ground->setCastShadows(false);

		//FILE *fp = fopen("output.txt", "a+b");
        //if( !fp )
            //return;

		try {
			spMMASS mmass = CMMASS::getSingleton();
			spMASS mass = mmass->createMASS("ground");
			spSpace space = mass->createSpace();
			for (int i = 1; i <= NUMROWS; i++) {
				for (int j = 1; j <= NUMCOLS; j++) {
					char name[6];
					sprintf( name, "%dx%d", i, j);
					spSite site = space->createSite(string(name));

					if (j != 1) {
						char left_name[6];
						sprintf(left_name, "%dx%d", i, (j-1));

						spSite left_site = space->getSite(string(left_name));
						site->addAdjacentSite(left_site);
					}

					if (i != 1) {
						char top_name[6];
						sprintf(top_name, "%dx%d", (i-1), j);

						spSite top_site = space->getSite(string(top_name));
						site->addAdjacentSite(top_site);
					}
				}
			}
			space->finish();

			// Creating "sheeps"
			AgentSheep::AgentSheepFactory factory;
			for (int i = 1, j = 1, k = 1; i <= 2; i++, k++) {
				char name[8];
				sprintf( name, "sheep%d", i );

				Entity *ent = mSceneMgr->createEntity(name, "ball.mesh");
				ent->setMaterialName("Examples/Blue");
				ent->setCastShadows(true);
				SceneNode *node = mSceneMgr->getRootSceneNode()->createChildSceneNode(string(name) + "node", Vector3((10*(k + (NUMROWS/2 - 2)))-5,5,(10*j)-5));
				node->attachObject(ent);

				const string siteName(_getSiteName(node->getPosition()));
				spSite site = space->getSite(string(siteName));

				spAgent agent = mass->createAgent(factory, string(name));
				agent->setSite(site);
				AgentSheep *sheep = dynamic_cast<AgentSheep*>(agent.get());

				entities[ent->getName()].mEntity = ent;
				entities[ent->getName()].mNode = node;
				entities[ent->getName()].mDestination = node->getPosition();
				entities[ent->getName()].state = sheep->getState();

				if ((i % 5) == 0) {
					j++;
					k = 0;
				}
			}
			mass->finish();
			mmass->finish();
			mmass->initiate();
		} catch (Space::logic_error &e) {
			//fprintf(fp, "%s\n", e.what()); fflush(fp);
		}
		//fclose(fp);

		mCamera->setPosition( 90.0f, 280.0f, 535.0f );
        mCamera->pitch( Degree(-30.0f) );
        mCamera->yaw( Degree(-15.0f) );
	}

	void createFrameListener(void) {
		mFrameListener = new MoveDemoListener(mWindow, mCamera, entities);
		mFrameListener->showDebugOverlay(true);
		mRoot->addFrameListener(mFrameListener);
	}
};

#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
#define WIN32_LEAN_AND_MEAN
#include "windows.h"
#endif

#ifdef __cplusplus
	extern "C" {
#endif

#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
		INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR strCmdLine, INT )
#else
		int main(int argc, char *argv[])
#endif
		{
			// Create application object
			MoveDemoApplication app;

			SET_TERM_HANDLER;

			try {
				app.go();
			} catch( Ogre::Exception& e ) {
#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
				MessageBox( NULL, e.getFullDescription().c_str(), "An exception has occured!", MB_OK | MB_ICONERROR | MB_TASKMODAL);
#else
				std::cerr << "An exception has occured: " <<
					e.getFullDescription().c_str() << std::endl;
#endif
			}

			return 0;
		}

#ifdef __cplusplus
	}
#endif
