#include "TestSpace.h"
#include "../MMASSdll/Site.h"
#include "../MMASSdll/MMASS.h"
using namespace MMASS;

void TestSpace::run() {
	creatingSiteTests();
	gettingSiteTests();
	creatingSiteWithAdjSitesTests();
}

void TestSpace::creatingSiteTests() {
	spMMASS mmass = CMMASS::getSingleton();
	spSpace space;

	try {
		space = mmass->getMASS("mass1")->getSpace();
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Creating a site using a space. */
		spSite site = space->createSite("site1.1");

		test_(site);
		test_(site->getSelf() == site);
		test_(site->getName() == "site1.1");
		test_(!site->hasAdjacentSites());
		test_(!site->hasAgent());
		test_(site->getSpace() == space);
		test_(space->numOfSites() == 1);
		test_(space->hasSite("site1.1"));
		test_(space->hasSite(site));
		test_(!space->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create two or more sites with the
			same name using the same space. */
		test_(space->hasSite("site1.1"));
		spSite site = space->createSite("site1.1");

		fail_("A site with this name already exists in this "
			"space. It should have thrown an exception.");
	} catch (Space::invalid_argument&) {
		succeed_();
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* But you can create sites with the same name
			if they are in different spaces. */
		spSpace space2 = mmass->getMASS("mass2")->getSpace();

		test_(space != space2);
		spSite site = space2->createSite("site1.1");

		test_(site);
		test_(site->getSelf() == site);
		test_(site->getName() == "site1.1");
		test_(!site->hasAdjacentSites());
		test_(!site->hasAgent());
		test_(site->getSpace() == space2);
		test_(space2->numOfSites() == 1);
		test_(space2->hasSite("site1.1"));
		test_(space2->hasSite(site));
		test_(!space2->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Creating one more site. */
		spSite site = space->createSite("site1.2");

		test_(site);
		test_(site->getSelf() == site);
		test_(site->getName() == "site1.2");
		test_(!site->hasAdjacentSites());
		test_(!site->hasAgent());
		test_(site->getSpace() == space);
		test_(space->numOfSites() == 2);
		test_(space->hasSite("site1.2"));
		test_(space->hasSite(site));
		test_(!space->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Creating some more sites. */
		space->createSite("site1.3");
		space->createSite("site1.4");
		space->createSite("site1.5");

		test_(space->numOfSites() == 5);

		space = mmass->getMASS("mass4")->getSpace();
		space->createSite("site4.1");

		test_(space->numOfSites() == 1);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}
}

void TestSpace::gettingSiteTests() {
	spMMASS mmass = CMMASS::getSingleton();
	spSpace space;

	try { /* Can't get a site from a MASS that doesn't have
		  any sites. */
		space = mmass->getMASS("mass5")->getSpace();
		test_(!space->hasSite());
		spSite site = space->getSite("site1.1");

		fail_("Space doesn't have a site. "
			"It should have thrown an exception.");
	} catch (Space::logic_error&) {
		succeed_();
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't get a site that doesn't exist in the
			space. */
		space = mmass->getMASS("mass1")->getSpace();
		test_(space->hasSite());
		test_(!space->hasSite("whole?"));
		spSite site = space->getSite("whole?");

		fail_("Space doesn't have a site. "
			"It should have thrown an exception.");
	} catch (Space::invalid_argument&) {
		succeed_();
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Getting a site. */
		spSite site = space->getSite("site1.2");

		test_(site);
		test_(site->getSelf() == site);
		test_(site->getName() == "site1.2");
		test_(!site->hasAdjacentSites());
		test_(!site->hasAgent());
		test_(site->getSpace() == space);
		test_(space->numOfSites() == 5);
		test_(space->hasSite("site1.2"));
		test_(space->hasSite(site));
		test_(!space->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Getting sites and comparing them. */
		spSite site11 = space->getSite("site1.1");
		spSite site12 = space->getSite("site1.2");
		spSite site21 = mmass->getMASS("mass2")->getSpace()->getSite("site1.1");

		test_(site11 != site12);
		test_(site11 != site21);
		test_(site12 != site21);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}
}

void TestSpace::creatingSiteWithAdjSitesTests() {
	spMMASS mmass = CMMASS::getSingleton();
	spSpace space;

	try {
		space = mmass->getMASS("mass1")->getSpace();
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Creating a site adjacent to a set of
			other sites. */
		spSite_set set;
		set.insert(space->getSite("site1.3"));
		set.insert(space->getSite("site1.4"));
		set.insert(space->getSite("site1.5"));

		spSite site = space->createSite("site1.6", set);

		test_(site);
		test_(site->getSelf() == site);
		test_(site->getName() == "site1.6");
		test_(site->numOfAdjacentSites() == 3);
		test_(!site->isAdjacentTo("site1.2"));
		test_(site->isAdjacentTo("site1.3"));
		test_(site->isAdjacentTo("site1.4"));
		test_(site->isAdjacentTo("site1.5"));
		test_(!site->isAdjacentTo("site1.6"));
		test_(!space->getSite("site1.2")->isAdjacentTo(site));
		test_(space->getSite("site1.3")->isAdjacentTo(site));
		test_(space->getSite("site1.4")->isAdjacentTo(site));
		test_(space->getSite("site1.5")->isAdjacentTo(site));
		test_(!space->getSite("site1.6")->isAdjacentTo(site));
		test_(space->getSite("site1.3")->numOfAdjacentSites() == 1);
		test_(space->getSite("site1.4")->numOfAdjacentSites() == 1);
		test_(space->getSite("site1.5")->numOfAdjacentSites() == 1);
		test_(!site->hasAgent());
		test_(site->getSpace() == space);
		test_(space->numOfSites() == 6);
		test_(space->hasSite("site1.6"));
		test_(space->hasSite(site));
		test_(!space->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Creating a site adjacent to a empty
		  set of sites (i.e., adjacent to any site).
		  It is equivalent to use the method
		  'createSite(aString)'. */
		spSite_set set;
		
		spSite site = space->createSite("site1.7", set);

		test_(site);
		test_(site->getSelf() == site);
		test_(site->getName() == "site1.7");
		test_(!site->hasAdjacentSites());
		test_(!site->hasAgent());
		test_(site->getSpace() == space);
		test_(space->numOfSites() == 7);
		test_(space->hasSite("site1.7"));
		test_(space->hasSite(site));
		test_(!space->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create a site adjacent to a set
			containing a NULL site. */
		spSite_set set;
		spSite nullSite;
		set.insert(space->getSite("site1.1"));
		set.insert(space->getSite("site1.2"));
		set.insert(nullSite);

		spSite site = space->createSite("site1.8", set);

		fail_("The set has a null site. "
			"It should have thrown an exception.");
	} catch (Space::invalid_argument&) {
		succeed_();

		test_(!space->getSite("site1.1")->hasAdjacentSites());
		test_(!space->getSite("site1.2")->hasAdjacentSites());
		test_(!space->hasSite("site1.8"));
		test_(space->numOfSites() == 7);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create a site adjacent to a set
			containing a site of another space, i.e.,
			a site can't be adjacent to a site of
			different space. */
		spSpace space2 = mmass->getMASS("mass2")->getSpace();
		spSite_set set;

		set.insert(space->getSite("site1.1"));
		set.insert(space2->getSite("site1.1"));

		spSite site = space->createSite("site1.8", set);

		fail_("The set has a site from a different space. "
			"It should have thrown an exception.");
	} catch (Space::invalid_argument&) {
		succeed_();
		spSpace space2 = mmass->getMASS("mass2")->getSpace();

		test_(!space->getSite("site1.1")->hasAdjacentSites());
		test_(!space2->getSite("site1.1")->hasAdjacentSites());
		test_(!space->hasSite("site1.8"));
		test_(space->numOfSites() == 7);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}
}

void TestSpaceAgain::run() {
	buildGraphs();
	finishTests();
	sitesDistancesTests();
}

void TestSpaceAgain::buildGraphs() {
	/* This method isn't a test. It just
	build a graph in each space, so we can
	test the ready() method. */
	spMMASS mmass = CMMASS::getSingleton();

	try { /* mass1 has a space with a connected graph. */
		spSpace space1 = mmass->getMASS("mass1")->getSpace();

		spSite site = space1->getSite("site1.2");
		site->addAdjacentSite(space1->getSite("site1.7"));

		site = space1->createSite("site1.8");
		site->addAdjacentSite(space1->getSite("site1.7"));

		site = space1->createSite("site1.9");
		site->addAdjacentSite(space1->getSite("site1.8"));
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* mass2 will have a graph with two connected
		  components (it won't be a connected graph). */
		spSpace space2 = mmass->getMASS("mass2")->getSpace();

		spSite site = space2->createSite("site2.1");
		site->addAdjacentSite(space2->createSite("site2.2"));

		site = space2->getSite("site1.1");
		site->addAdjacentSite(space2->createSite("site1.2"));
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* mass3 will have a graph with one site. */
		spSpace space3 = mmass->getMASS("mass3")->getSpace();

		space3->createSite("site3.1");
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	/* mass4 will have a graph with no sites. */
}

void TestSpaceAgain::finishTests() {
	spMMASS mmass = CMMASS::getSingleton();

	try { /* The space from mass1 must be ok. */
		spSpace space1 = mmass->getMASS("mass1")->getSpace();
		space1->finish();

		test_(space1->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't finish a space with more then one connected
		  component. */
		spSpace space2 = mmass->getMASS("mass2")->getSpace();
		space2->finish();

		fail_("The mass2 doesn't have a connected graph."
			"It should have thrown an exception.");
	} catch (Space::logic_error&) {
		succeed_();

		spSpace space2 = mmass->getMASS("mass2")->getSpace();
		test_(!space2->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* A graph with one node is a connected graph.
		  So, we can finish the space from mass3. */
		spSpace space3 = mmass->getMASS("mass3")->getSpace();
		space3->finish();

		test_(space3->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Same as before. */
		spSpace space4 = mmass->getMASS("mass4")->getSpace();
		space4->finish();

		test_(space4->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Finish a space more than one time doesn't have effects. */
		spSpace space4 = mmass->getMASS("mass4")->getSpace();
		test_(space4->isReady());

		space4->finish();
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't finish a space with no sites. */
		spSpace space5 = mmass->getMASS("mass5")->getSpace();
		space5->finish();

		fail_("The mass5 doesn't have any site."
			"It should have thrown an exception.");
	} catch (Space::logic_error&) {
		succeed_();

		spSpace space5 = mmass->getMASS("mass5")->getSpace();
		test_(!space5->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create a site with a finished space. */
		spSpace space1 = mmass->getMASS("mass1")->getSpace();
		test_(space1->numOfSites() == 9);
		test_(space1->isReady());

		spSite site = space1->createSite("site1.10");

		fail_("This space is finished. "
			"It should have thrown an exception.");
	} catch (Space::logic_error&) {
		succeed_();

		spSpace space1 = mmass->getMASS("mass1")->getSpace();
		test_(space1->numOfSites() == 9);
		test_(space1->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't create a site with a finished space. */
		spSpace space1 = mmass->getMASS("mass1")->getSpace();
		test_(space1->numOfSites() == 9);
		test_(space1->isReady());

		spSite_set set;
		spSite site = space1->createSite("site1.10", set);

		fail_("This space is finished. "
			"It should have thrown an exception.");
	} catch (Space::logic_error&) {
		succeed_();

		spSpace space1 = mmass->getMASS("mass1")->getSpace();
		test_(space1->numOfSites() == 9);
		test_(space1->isReady());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}
}

void TestSpaceAgain::sitesDistancesTests() {
	spSite nullSite;
	spSpace space = CMMASS::getSingleton()->getMASS("mass1")->getSpace();

	try {
		spSite_map::const_iterator site_i, s_end = space->endSites();
		for (site_i = space->beginSites(); site_i != s_end; ++site_i) {
			spSite previousSite, site = (*site_i).second;
			queue<const string> sitesInOrder = site->getSitesInDistanceOrder();
			while (!sitesInOrder.empty()) {
				const string siteName = sitesInOrder.front();
				sitesInOrder.pop();
				spSite currentSite = space->getSite(siteName);
				test_(site->getDistanceOf(currentSite) == currentSite->getDistanceOf(site));
				/*printf("\n%s from %s = %lu, %s from %s = %lu\n",
					site->getName().c_str(), currentSite->getName().c_str(), site->getDistanceOf(currentSite),
					currentSite->getName().c_str(), site->getName().c_str(), currentSite->getDistanceOf(site));*/

				if (previousSite != nullSite)
					test_(site->getDistanceOf(previousSite) <= site->getDistanceOf(currentSite));

				previousSite.reset();
				previousSite = currentSite;
			}
			/*printf("\n%s from %s = %lu\n", site->getName().c_str(), site->getName().c_str(), site->getDistanceOf(site));*/
		}
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}
}